            .LF     irtoslink.lst
;-------------------------------------------------------------------
;
; Project : irtoslink
; File    : irtoslink
; Author  : San Bergmans
; Date    : June 2012
;
; The function of this device is to relay IR signals from a receiver
; which is located in one room, to an IR-LED located in another
; room. And it will select one of 4 TOSLINK inputs and connect that
; to one TOSLINK output. The input can be selected with one push
; button on the front panel and by audio IR commands.
;
;-------------------------------------------------------------------

            .CR     AVR
            .FA     MEGA
            .TF     irtoslink.hex,int

;-------------------------------------------------------------------
; Declarations
;-------------------------------------------------------------------

PINB        .EQ     $03         Port B input
DDRB        .EQ     $04         Port B data direction
PORTB       .EQ     $05         Port B data register
PINC        .EQ     $06         Port C input
DDRC        .EQ     $07         Port C data direction
PORTC       .EQ     $08         Port C data register
PIND        .EQ     $09         Port D input
DDRD        .EQ     $0A         Port D data direction
PORTD       .EQ     $0B         Port D data register

; Port definitions
; B0 = IR-LED ouptut (active low)
; B1 = IR receiver input (active low)
; B2 = IR feedback LED (active low)
; D0..D1 = Toslink input select
; D2 = Toslink disable
; D3 = Input 1 LED
; D4 = Input 2 LED
; D5 = Input 3 LED
; D6 = Input 4 LED
; D7 = Input select button

; Timer 2 registers
TCCR2A      .EQ     $B0         Timer control register A
TCCR2B      .EQ     $B1         Timer control register B
TCNT2       .EQ     $B2         Timer/counter 2 register
OCR2A       .EQ     $B3         Output compare register A
OCR2B       .EQ     $B4         Output compare register B

; Interrupt registers
TIFR2       .EQ     $17         Timer interrupt flag register
TIMSK2      .EQ     $70         Timer interrupt mask register
SMCR        .EQ     $33         Sleep mode control register

SPL         .EQ     $3D         Stack pointer low
SPH         .EQ     $3E         Stack pointer high
SREG        .EQ     $3F         System register

RAMEND      .EQ     $4FF        Last byte of RAM memory

;-------------------------------------------------------------------
; RAM declarations (Everything fits in register bank)
;-------------------------------------------------------------------

KEY_COUNT   .EQ     $00         Key is pressed if not 0
TOSLINK_IN  .EQ     $01         Currently selected toslink input
MUTE_DELAY  .EQ     $02,$03     Mute delay (steps of 26.3 us)
IR_COMMAND  .EQ     $04,$05     Received IR message (command)
IR_ADDRESS  .EQ     $06,$07     Received IR message (address)
LED_DELAY   .EQ     $08         IR receive feedback LED flash cntr
IR_COUNT    .EQ     24,25       IR cycle counter (1 cycle is 26.3us)
IR_STATE    .EQ     30,31       IR state machine pointer (Z reg)

;-------------------------------------------------------------------
; Constants
;-------------------------------------------------------------------

PRE0_MIN    .EQ     274         Pre pulse low minimum time (-20%)
PRE0_MAX    .EQ     411         Pre pulse low maximum time (+20%)
PRE1_MIN    .EQ     137         Pre pulse high minimum time (-20%)
PRE1_MAX    .EQ     205         Pre pulse high maximum time (+20%)
BIT0_MIN    .EQ     17          Bit pulse low minimum time (-20%)
BIT0_MAX    .EQ     26          Bit pulse low maximum time (+20%)
BIT1_HALF   .EQ     57          Bit pulse half between 0 and 1
BIT1_MAX    .EQ     102         Bit pulse high maximum time (+20%)
IR_GAP      .EQ     190         Minimum end gap time (5 ms)

BOSE_ADDR   .EQ     $4BBA       Bose address (Humax code 1229)
BOSE_1      .EQ     $59         Bose select input 1 key
BOSE_2      .EQ     $53         Bose select input 2 key
BOSE_3      .EQ     $FF         Bose select input 3 key
BOSE_4      .EQ     $0E         Bose select input 4 key

;-------------------------------------------------------------------
; Vector space
;-------------------------------------------------------------------

RESET       RJMP    INIT        Reset vector
INT0        RETI                External Interrupt 0
INT1        RETI                External Interrupt 1
PCINT0      RETI                Pin Change Interrupt 0
PCINT1      RETI                Pin Change Interrupt 1
PCINT2      RETI                Pin Change Interrupt 2
WDT         RETI                Watchdog time-out interrupt
TIMER2COMPA RJMP    T2COMPA     Timer 2 compare match A
TIMER2COMPB RJMP    T2COMPB     Timer 2 compare match B
TIMER2OVF   RETI                Timer 2 overflow
TIMER1CAPT  RETI                Timer 1 capture event
TIMER1COMPA RETI                Timer 1 compare match A
TIMER1COMPB RETI                Timer 1 compare match B
TIMER1OVF   RETI                Timer 1 overflow
TIMER0COMPA RETI                Timer 0 compare match A
TIMER0COMPB RETI                Timer 0 compare match B
TIMER0OVF   RETI                Timer 0 overflow
SPISTC      RETI                SPI serial transfer complete
USARTRX     RETI                USART RX complete
USARTUDRE   RETI                USART data register empty
USARTTX     RETI                USART TX complete
ADC         RETI                ADC conversion complete
EEREADY     RETI                EEPROM ready
ANALOGCOMP  RETI                Analog comparator
TWI         RETI                2-wire serial interface
SPMREADY    RETI                Store program memory ready

;-------------------------------------------------------------------
; T2COMPA   Timer interrupt, every 26.25 us, main timing loop
;-------------------------------------------------------------------

T2COMPA     SBIS    PINB,1      Turn IR LED on if input is low
            CBI     PORTB,0
            
            MOV     R9,SREG     Save status register

            ICALL               Call to current IR state machine
            CALL    KEY_IN      Read input select key
            CALL    MUTE_OFF    Clear mute after delay
            CALL    FLASH_LED   Flash IR receive feedback LED
            
            MOV     SREG,R9     Restore status register

            RETI

;-------------------------------------------------------------------
; T2COMPB   Timer interrupt, at 1/3rd of main timing loop
;           All we need to do here is to turn off the IR-LED
;-------------------------------------------------------------------

T2COMPB     SBI     PORTB,0     Turn IR-LED off
            RETI                Done!

;-------------------------------------------------------------------
; INIT      Get the system up and running
;-------------------------------------------------------------------

INIT        ;                   Initialize stack
            LDI     R16,#RAMEND
            OUT     SPL,R16
            LDI     R16,/RAMEND
            OUT     SPH,R16
            
            ;                   Define inputs and outputs                
            LDI     R16,#%1111.1101 Port B1 is an input, the rest
            OUT     DDRB,R16         are outputs (inputs souldn't
            OUT     PORTB,R16         be allowed to float)
            LDI     R16,#%1111.1111 All bits of port C are outputs
            OUT     PORTC,R16        to avoid floaating inputs
            OUT     DDRC,R16
            LDI     R16,#%0111.1111
            OUT     DDRD,R16    Port D7 is an input, rest is output
            
            LDI     IR_STATE,#IRS_IDLE  Init IR state machine
            LDI     IR_STATE+1,/IRS_IDLE
            
            ;                   Initialize input selector
            LDI     R18,#0      Select input 1
            MOV     KEY_COUNT,R16
            MOV     TOSLINK_IN,R18
            INC     TOSLINK_IN  Make old input value invalide
            CALL    SET_INPUT   Set input and LEDs
            
            ;                   Initialize Timer/Counter 2
            ;                   This timer is used to generate
            ;                    the 38kHz modulation of the
            ;                    IR output
            LDI     R27,#0      Address extended I/O (high byte = 0)
            LDI     R26,#TCCR2A Point to extended I/O register
            LDI     R16,#%0000.0010  Select CTC mode
            ST      X+,R16          (TCCR2A)
            LDI     R16,#%0000.0001  Select systemclock / 1
            ST      X+,R16          (TCCR2B)
            INC     R26             (skip TCNT2)
            LDI     R16,#210    Set max count to 8Mhz / 38kHz
            ST      X+,R16          (OCR2A)
            LDI     R16,#70     Set 1/3rd of max (pulse width)
            ST      X+,R16          (OCR2B)
            LDI     R26,#TIMSK2 Point to extended I/O register
            LDI     R16,#%0000.0110 Allow interrupts on both compares
            ST      X,R16
            
            SEI                     Start the interrupts
            LDI     R16,#1          Set sleep mode register
            OUT     SMCR,R16
            
;-------------------------------------------------------------------
; MAIN      Main program loop
;-------------------------------------------------------------------

.4EVER      SLEEP                   Nothing to do here, sleep until
            JMP     .4EVER           woken by an interrupt

;-------------------------------------------------------------------
; KEY_IN    Read input select key and select next input if pressed
;-------------------------------------------------------------------

KEY_IN      SBIS    PIND,7      Skip next instrucion if key released
            RJMP    .KEY_DOWN   Handle key down!
            
            ;                   Key is released
            LDI     R18,#0      Clear key counter
            MOV     KEY_COUNT,R18  and we're done
            RET
            
.KEY_DOWN   INC     KEY_COUNT   Increment key counter
            BRNE    .NO_OVFL    No overflow!
            DEC     KEY_COUNT   Don't allow overflow of counter
.DONE       RET
            
.NO_OVFL    LDI     R18,#190    Accept key after 5ms
            CP      R18,KEY_COUNT
            BRNE    .DONE       We're done!

            MOV     R18,TOSLINK_IN   Increment input
            INC     R18

; This routine rolls through to the next routine!!!

;-------------------------------------------------------------------
; SET_INPUT Set Toslink input
; Input :   R18 holds input to select (0..3)
;-------------------------------------------------------------------

SET_INPUT   ANDI    R18,#%0000.0011 Bring value within limits
            CP      R18,TOSLINK_IN
            BREQ    .RET            This input is already selected!
            
            LDI     R19,#0
            MOV     TOSLINK_IN,R18  Set required input
            PUSH    R30             Save Z (normally holds IR state)
            PUSH    R31
            LDI     R30,#.TABLE*2   Select proper LED and input
            LDI     R31,/.TABLE*2     selector
            ADD     R30,R18
            ADC     R31,R19    
            LPM     R18,Z
            OUT     PORTD,R18
            LDI     R18,/3800       Set mute delay to 100 ms
            MOV     MUTE_DELAY+1,R18     ( 100ms / 26.3 us)
            LDI     R18,#3800
            MOV     MUTE_DELAY,R18
            POP     R31             Restore IR state
            POP     R30
.RET        RET

.TABLE      .DA     #%0111.0100     Input 1
            .DA     #%0110.1101     Input 2
            .DA     #%0101.1110     Input 3
            .DA     #%0011.1111     Input 4
            .EV
            
;-------------------------------------------------------------------
; MUTE_OFF  Wait 100 ms before turning the Mute off
;-------------------------------------------------------------------

MUTE_OFF    ;               See if mute delay is 0 already
            MOV     R18,MUTE_DELAY
            OR      R18,MUTE_DELAY+1
            BREQ    .RET    It is!

            LDI     R18,#0  See if we need to borrow from high byte
            CP      R18,MUTE_DELAY
            BRNE    .NOBORROW
            DEC     MUTE_DELAY+1 Borrow from high byte
.NOBORROW   DEC     MUTE_DELAY   Decrement low byte too

            MOV     R18,MUTE_DELAY   Counter 0 now?
            OR      R18,MUTE_DELAY+1
            BRNE    .RET             Not yet!
            
            CBI     PORTD,2     Clear Mute
.RET        RET

;-------------------------------------------------------------------
; FLASH_LED  Flash the IR receive feedback LED if input is low
;-------------------------------------------------------------------

FLASH_LED   LDI     R18,#0          Is the delay counter 0?
            CP      LED_DELAY,R18
            BRNE    .LED_ON         Nope! LED must be on
            CBI     PORTB,2         Turn LED off
            RET

.LED_ON     DEC     LED_DELAY       Decrement delay counter
            SBI     PORTB,2         Turn LED on
            RET

;-------------------------------------------------------------------
; IRS_IDLE  IR state machine, idle state, wait for lead pulse
;-------------------------------------------------------------------

IRS_IDLE    SBIC    PINB,1      Skip if input is low
            RET                 Input is still high, nothing to do!

            LDI     IR_COUNT,#0         Clear pulse length timer 
            LDI     IR_COUNT+1,#0
            MOV     IR_ADDRESS,IR_COUNT Clear IR message
            MOV     IR_ADDRESS+1,IR_COUNT
            MOV     IR_COMMAND,IR_COUNT
            LDI     R18,#%1000.0000     Set end flag
            MOV     IR_COMMAND+1,R18

            LDI     IR_STATE,#IRS_PRE0 Next state, time pre pulse
            LDI     IR_STATE+1,/IRS_PRE0
            RET

;-------------------------------------------------------------------
; IRS_PRE0  IR state machine, pre pulse is low, time its length
;-------------------------------------------------------------------

IRS_PRE0    INC     IR_COUNT    Increment pulse length timer
            BRNE    .NOC        No carry!
            INC     IR_COUNT+1

.NOC        SBIC    PINB,1      Is input still low, or gone high?
            RJMP    .HIGH       It's gone high!

            CPI     IR_COUNT+1,#0
            BREQ    .RET        Pulse can't be too long yet!
            CPI     IR_COUNT,#PRE0_MAX-256
            BRCC    IRS_ERR0        Pulse too long!
.RET        RET

.HIGH       CPI     IR_COUNT+1,#1
            BRNE    IRS_ERR1    Pulse was by far nog long enough!
            CPI     IR_COUNT,#PRE0_MIN-256
            BRCS    IRS_ERR1    Pulse too short, ignore!

            LDI     R18,#255        Flash the IR feedback light
            MOV     LED_DELAY,R18

            LDI     IR_COUNT,#0         Clear timer for next state
            LDI     IR_STATE,#IRS_PRE1  Next state measure
            LDI     IR_STATE+1,/IRS_PRE1   pre pulse gap
            RET

;-------------------------------------------------------------------
; IRS_PRE1  IR sate machine, pre pulse is high, time its length
;-------------------------------------------------------------------

IRS_PRE1    INC     IR_COUNT    Increment pulse length timer
            SBIS    PINB,1      Is input still high, or gone low?
            RJMP    .LOW        It's gone low!

            CPI     IR_COUNT,#PRE1_MAX
            BRCC    IRS_ERR1    Pulse too long high!
.RET        RET

.LOW        CPI     IR_COUNT,#PRE1_MIN
            BRCS    IRS_ERR0        Pulse was too short, ignore!
        
            LDI     IR_COUNT,#0         Clear timer for next state
            LDI     IR_STATE,#IRS_BIT0  Next state is bit low
            LDI     IR_STATE+1,/IRS_BIT0
            RET

;-------------------------------------------------------------------
; IRS_ERR0  IR state machine, error state, wait for input rise
;-------------------------------------------------------------------

IRS_ERR0    LDI     IR_STATE,#IRS_ERR0  Set state to self
            LDI     IR_STATE+1,/IRS_ERR0

            SBIS    PINB,1      Skip next if input is high
            RET                 Input is still low!

; This routine rolls over into the next, don't insert code here.

;-------------------------------------------------------------------
; IRS_ERR1  IR state machine, error state, input is high already
;-------------------------------------------------------------------

IRS_ERR1    LDI     IR_STATE,#IRS_IDLE   Return to idle state
            LDI     IR_STATE+1,/IRS_IDLE
            RET

;-------------------------------------------------------------------
; IRS_BIT0  IR state machine, receiving a bit while input is low
;-------------------------------------------------------------------

IRS_BIT0    INC     IR_COUNT    Increment pulse length timer
            SBIC    PINB,1      Is input still low, or gone high?
            RJMP    .HIGH       It's gone high!
            
            CPI     IR_COUNT,#BIT0_MAX
            BRCC    IRS_ERR0        Pulse too long low!
            RET

.HIGH       CPI     IR_COUNT,#BIT0_MIN
            BRCS    IRS_ERR1        Pulse was too short, ignore!

;                                   Don't clear length timer!
            LDI     IR_STATE,#IRS_BIT1  Next state is bit high
            LDI     IR_STATE+1,/IRS_BIT1
            RET

;-------------------------------------------------------------------
; IRS_BIT1  IR state machine, receiving a bit while input is high
;-------------------------------------------------------------------

IRS_BIT1    INC     IR_COUNT    Increment pulse length timer
            SBIS    PINB,1      Is input still high, or gone low?
            RJMP    .LOW        It's gone low!
            
            CPI     IR_COUNT,#BIT1_MAX
            BRCC    IRS_ERR1        Pulse too long high!
            RET

.LOW        LDI     R18,#BIT1_HALF  Was it a 0 ( < 1.5 bit time)
            CP      R18,IR_COUNT     or was it a 1 ( > 1.5 bit time)
            ROR     IR_COMMAND+1    Carry represented bit
            ROR     IR_COMMAND
            ROR     IR_ADDRESS+1
            ROR     IR_ADDRESS
            BRCS    .DONE           All bits received
            
            LDI     IR_COUNT,#0     Restart IR timer for next bit
            LDI     IR_STATE,#IRS_BIT0      Next stop, bit low!
            LDI     IR_STATE+1,/IRS_BIT0
            RET

.DONE       LDI     IR_COUNT,#0     Restart IR timer for end pulse
            LDI     IR_STATE,#IRS_END0      Next stop, end pulse
            LDI     IR_STATE+1,/IRS_END0
            RET

;-------------------------------------------------------------------
; IRS_END0  IR state machine, end bit is 0, wait for end bit to end
;-------------------------------------------------------------------

IRS_END0    INC     IR_COUNT    Increment pulse length timer
            SBIC    PINB,1      Is input still low, or gone high?
            RJMP    .HIGH       It's gone high!
            
            CPI     IR_COUNT,#BIT0_MAX
            BRCC    IRS_ERR0    Pulse too long low!
            RET

.HIGH       CPI     IR_COUNT,#BIT0_MIN
            BRCS    IRS_ERR1    Pulse was too short, ignore!

            MOV     IR_COUNT,IR_COMMAND Does checksum match?
            OR      IR_COUNT,IR_COMMAND+1
            INC     IR_COUNT            Result is 0 for next state
            BRNE    IRS_ERR1            Nope!
            LDI     IR_STATE,#IRS_END1  Next state is end high
            LDI     IR_STATE+1,/IRS_END1
            RET

;-------------------------------------------------------------------
; IRS_END1  IR state machine, received entire message
;-------------------------------------------------------------------

IRS_END1    INC     IR_COUNT    Increment length timer
            SBIS    PINB,1      Is input still high, or gone low?
            RJMP    IRS_ERR0    It's gone low, that can't be right!
            
            CPI     IR_COUNT,#IR_GAP Is gap long enough now?
            BRCS    .RET        Not yet!
       
            LDI     R18,#BOSE_ADDR  Is it the correct address?
            CP      R18,IR_ADDRESS
            BRNE    .NOP            No!, Ignore command
            LDI     R18,/BOSE_ADDR
            CP      R18,IR_ADDRESS+1
            BRNE    .NOP            No!, Ignore command

            LDI     R18,#0          Begin with input 1
            MOV     R19,IR_COMMAND
            CPI     R19,#BOSE_1
            BREQ    .DO_IT          It's input 1!
            INC     R18
            CPI     R19,#BOSE_2
            BREQ    .DO_IT          It's input 2!
            INC     R18
            CPI     R19,#BOSE_3
            BREQ    .DO_IT          It's input 3!
            INC     R18
            CPI     R19,#BOSE_4
            BRNE    .NOP            Not the right key, ignore it!

.DO_IT      RCALL   SET_INPUT       Select right input
            
.NOP        LDI     IR_STATE,#IRS_IDLE  Return to idle stat now
            LDI     IR_STATE+1,/IRS_IDLE
.RET        RET

